#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <stdarg.h>
#include <float.h>
#include <cpgplot.h>

/***************************************************************
 *
 *   LIMITS  ---  MODIFY TO SUIT YOUR PROBLEM
 *
 *   ORDER IS  Gamma, K1, e, omega, P(eriod), T(ime), K1/K2 
 *
 *   LOWER LIMITS (allow omega < 0.0 to improve convergence)
 */
double lim_low[7]  = { -1000.0, -1000.0,  0.0,  -7.0,   0.0,   0.0,   0.01 };
/*
 *   UPPER LIMITS
 */
double lim_hig[7]  = {  1000.0,  1000.0, 0.99,   7.0, 100.0, 100.0,  100.0 };
/*
 *
 *
 ***************************************************************/
/*
=======================================================================

Copyright (C) 2001-2004  Rainer Wichmann <rwichmann at la-samhna dot de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    In addition, as a special exception, Rainer Wichmann
    gives permission to link the code of this program with
    the PGPLOT library (or with modified versions of PGPLOT that use the
    same license as PGPLOT), and distribute linked combinations including
    the two.  You must obey the GNU General Public License in all
    respects for all of the code used other than PGPLOT.  If you modify
    this file, you may extend this exception to your version of the
    file, but you are not obligated to do so.  If you do not wish to
    do so, delete this exception statement from your version.

=======================================================================
*/

/* Changelog:
 * 
 * 1.3: fix DBL_EPSILON -> 20 * DBL_EPSILON in nu_of_t()
 *      fix initialisation of simplex in SimplexInit()
 *      fix initialisation of init vector for test data
 *      (bug report by Monique Spite and Claire Bentolila)
 */

/* penalty for out-of-bounds                                 
 */
#define PENALTY 1.0E37

#define MAX_ITER_SIMPLEX   200

typedef struct   BlahParStr {
  double   Lower;        /* upper limit                                    */
  double   Upper;        /* lower limit                                    */
  double   Range;        /* range                                          */
  char     Name[32];     /* name of parameter                              */
  double   t;            /* parameter temperature                          */
} ParStr;

/* max. number of free parameters
 */
#define SDIM  7

#define TWOPI  (2.0 * M_PI);
#define _(string) string
#define ON  1
#define OFF 0

int g_nparams = 6;
int verbose = 0;
int debug   = 0;
double SimplexTol = FLT_EPSILON;

/* default output device for pgplot
 */
char * Device = "/XSERVE";

struct data {
  size_t n;
  double * x;
  double * y;
  double * sigma;
  int    * flag;
};

int   
plot_result (float gamma, float K, float e, float omega, float P, 
	     float T, float K2,
	     int dn, double *dxi, double *dyi, double *dsi, int *dfi);

int
plot_statistics (float gamma, float K, float e, float omega, 
		 float P, float T, float K2,
		 int dn, double *dxi, double *dyi, double *dsi, int *dfi);

void out_of_range()
{
  static int oor = 0;
  if (oor == 1)
    return;
  fprintf(stderr, "***INFO*** This error indicates that your input is out of the\n");
  fprintf(stderr, "***INFO*** compiled-in limits (which are there to avoid that\n");
  fprintf(stderr, "***INFO*** the fit diverges). Edit the limits in line 21/25 of\n");
  fprintf(stderr, "***INFO*** velocity.c and recompile to fix this problem.\n");
  oor = 1;
  return;
}

/* Standard error handler
 */
void die (int line, char * fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "***ERROR***: Line %4d: ", line);
  vfprintf(stderr, fmt, ap);
  va_end(ap);

  exit (EXIT_FAILURE);
}


void * xalloc (size_t size)
{
  void * rv = malloc(size);
  if (rv == NULL)
    die(__LINE__, "Out of memory\n");
  return rv;
}


ParStr Param[SDIM] = {
  {   -1000.,      1000.,   0.0, "Gamma",                    1.0 }, 
  {       0.,      1000.,   0.0, "K1",                       1.0 }, 
  {      0.0,       0.99,   0.0, "e",                        1.0 }, 
  {      0.0,        7.0,   0.0, "omega",                    1.0 }, 
  {      0.0,       20.0,   0.0, "P",                        1.0 }, 
  {      0.0,      100.0,   0.0, "T",                        1.0 }, 
  {      0.5,        1.5,   0.0, "K1/K2",                    1.0 }, 
};

void SimplexSetVariables(gsl_vector * f, double x[], int * Flags_simplex)
{
  int  j;

  if (debug) { 
    fprintf(stderr, _("SimplexSetVariables: Vector\n") );
    for (j = 0; j < g_nparams; ++j) fprintf(stderr, " %9.4f", x[j]);
    fprintf(stderr, "\n");
    fprintf(stderr, _("Flags\n") );
    for (j = 0; j < g_nparams; ++j) fprintf(stderr, " %3d", Flags_simplex[j]);
    fprintf(stderr, "\n");
  }

  for (j=0; j < g_nparams; ++j) 
    {
      if (Flags_simplex[j] == ON) 
	gsl_vector_set(f, j, x[j]);
    }

  return;
}

int SimplexCheckVertex(gsl_vector * f, double X[], int * Flags_simplex)
{
  int          i;                        /* loop counter                    */

  for (i = 0; i < g_nparams; ++i) {
    if (Flags_simplex[i] == ON) {
      if (X[i] <= (Param[i].Lower-FLT_EPSILON) 
	  || X[i] >= (Param[i].Upper+FLT_EPSILON)) {
	if (verbose)
	  fprintf(stdout, 
		  "*** Out of range [%8.4g,%8.4g]: %8s, Value = %8.4g\n",
		  Param[i].Lower,
		  Param[i].Upper,
		  Param[i].Name,
		  X[i]);
	return(GSL_EDOM);
      }
    }
  }

  SimplexSetVariables(f, X, Flags_simplex);
  return (GSL_SUCCESS);
}
  

void SimplexInitRanges(gsl_vector * f, double * x_init)
{
  int i;

  for (i=0; i<g_nparams; ++i)
    gsl_vector_set(f, i, x_init[i]);

  for (i=0; i<g_nparams; ++i) {
    Param[i].Upper = lim_hig[i];
    Param[i].Lower = lim_low[i];
    Param[i].Range = Param[i].Upper - Param[i].Lower;
    Param[i].t     = 1.0;
  }


  return;
}


int SimplexInit(double Y[], double S[][SDIM], int * Flags_simplex, 
		double (*func)(const gsl_vector *, void *),
		gsl_vector * xr, void * params)
{
  int     i, j;                                  /* loop counter            */
  int     vertexnum;                             /* vertex counter          */
  int     VertexTest;                            /* exit status             */
  double  p, q;                                  /* for vertex definition   */
  double  FitDim;                                /* # params to fit         */
  double  X[SDIM] = { 0.0 };                     /* a vertex                */
  double  Var[SDIM+1] = { 0.0 };                 /* init deltas             */
  double  tmp;                                   /* temporary               */

  /* -------------------  initialize -------------------------------------  */

   FitDim = 0;
   for (j = 0; j < g_nparams; ++j) {
      S[0][j] = 0.0;
      if (Flags_simplex[j] == ON) {
	++FitDim;
	S[0][j] = gsl_vector_get (xr, j);
      }
      Var[j+1] = Param[j].Range / 10.;
   }

  tmp = sqrt(2.0*FitDim);
  p = (sqrt(FitDim+1.0) + FitDim -1.0) / tmp;
  q = (sqrt(FitDim+1.0) - FitDim) / tmp;

  /* -------------- generate init vertex ---------------------------------- */

  for (j = 0; j < g_nparams; ++j) X[j] = S[0][j];
  VertexTest = SimplexCheckVertex(xr, X, Flags_simplex);

  if (VertexTest > 0) {
    Y[0] = PENALTY;
  } else {
    SimplexSetVariables(xr, X, Flags_simplex);
    Y[0] = func(xr, params);
  }

  /* ---------------- define vertices of simplex -------------------------  */

  vertexnum = 0;
  for (i = 1; i < g_nparams+1; ++i) {

    if (Flags_simplex[i-1] == ON) {

      for (j = 0; j < g_nparams; ++j) {
	   S[i][j] = S[0][j];
	   if(Flags_simplex[j] == ON) 
	     {
	       /*
	       if (j != (i-1))
		 printf("[%d][%d] %f = %f + %f * %f\n", i, j,
			S[i][j] + q * Var[j+1], S[i][j], q, Var[j+1]);
	       else
		 printf("[%d][%d] %f = %f + %f * %f\n", i, j,
			S[i][j] + p * Var[j+1], S[i][j], p, Var[j+1]);
	       */

	       S[i][j] = S[i][j] + q * Var[j+1];
	     }
      }
       
      S[i][i-1] = S[0][i-1] + p * Var[i];
      for (j = 0; j < g_nparams; ++j) 
	X[j] = S[i][j];
      VertexTest = SimplexCheckVertex(xr, X, Flags_simplex);
      
      if (VertexTest != GSL_SUCCESS) {
	Y[i] = PENALTY;
	if (verbose) 
	  printf(" -- Vertex %3d  initialized (invalid state)\n", 
		 vertexnum);
      } else {
	SimplexSetVariables(xr, X, Flags_simplex);
	Y[i] = func (xr, params);
	if (verbose) 
	  printf(" -- Vertex %3d  initialized \n",
		 vertexnum);
      }
      ++vertexnum;
    } else {
      for (j = 0; j < g_nparams; ++j) S[i][j] = 0.0;
      Y[i] = 0.0;
    }
  }
  
  return(GSL_SUCCESS);
}


double SimplexFlow(double Factor, double S[][SDIM], 
		   double Y[], double Centre[], int Worst, 
		   int FitDim, int *ErrCode,	
		   int * Flags_simplex,
		   double (*func)(const gsl_vector *, void *),
		   gsl_vector * xr, void * params)
{
   int      j;                       /* loop counter                        */
   double   Y_Test;                  /* the merit                           */
   double   X_Test[SDIM];            /* the parameter vector                */
   int      VertexTest;              /* exit status subroutine              */

   *ErrCode = GSL_SUCCESS;
   
   /* -------- compute parameter vector ---------------------------------   */

   for (j = 0; j < g_nparams; ++j) {
       if (Flags_simplex[j] == ON) { 
	 X_Test[j] = Centre[j] 
	   + Factor * (S[Worst][j] - Centre[j]);
       } else { 
	 X_Test[j] = 0.0; 
       }
   }

   if (debug) { 
      fprintf(stderr, "Centre:" );
      for (j = 0; j < g_nparams; ++j) 
	fprintf(stderr, " %7.2f", Centre[j]);
      fprintf(stderr, "\n");
      fprintf(stderr, "Worst:" );
      for (j = 0; j < g_nparams; ++j) 
	fprintf(stderr, " %7.2f", S[Worst][j]);
      fprintf(stderr, "\n");
      fprintf(stderr, "New:" );
      for (j = 0; j < g_nparams; ++j) 
	fprintf(stderr, " %7.2f", X_Test[j]);
      fprintf(stderr, "\n");
   }

   /* -------- validate and compute     ---------------------------------   */

   VertexTest = SimplexCheckVertex(xr, X_Test, Flags_simplex);
   if (VertexTest != GSL_SUCCESS) {
     Y_Test = PENALTY;
     return(Y_Test);
   } else {
     SimplexSetVariables(xr, X_Test, Flags_simplex);
     if (debug) { 
       for (j = 0; j < g_nparams; ++j) 
	 printf(" %7.2f", X_Test[j]);
       printf("\n");
     }
     Y_Test = func(xr, params);
   }

 
   /* -------- update simplex           ---------------------------------   */

   if (Y_Test <= (Y[Worst]-FLT_EPSILON)) {
     Y[Worst] = Y_Test;
     for (j = 0; j < g_nparams; ++j) {
       if (Flags_simplex[j] == ON) { 
	 Centre[j]   = 
	   ((FitDim+1) * Centre[j] - S[Worst][j] + X_Test[j]) / (FitDim+1);
	 S[Worst][j] = X_Test[j];
       }
     }
   }

   return(Y_Test);
}



int Simplex( double * x_init,  int * Flags_simplex, double * chi_new,
	     double (*func)(const gsl_vector *, void *),
	     gsl_vector * xr, void * params)
{

   int         ErrCode = 0;             /* exit status subroutines     */
   double      Y[SDIM+1] = {0.0};       /* merit (result) vector       */
   double      X[SDIM] = {0.0};         /* some Vertex                 */
   double      BestEver[SDIM] = {0.0};  /* best vertex found           */
   double      BestChi = 1E38;          /* best chi found              */
   double      S[SDIM+1][SDIM];         /* The Simplex                 */
   double      Centre[SDIM] = {0.0};    /* Simplex Centre              */
   int         FitDim;                  /* Number of Fitted Parameters */
   int         Worst = 0;               /* Index of worst Vertex       */
   int         NextWorst = 0;           /* Index of second worst Vertex*/
   int         Best = 0;                /* Index of best Vertex        */

   double      Sum;                     /* temporary sum               */
   long        NumIter = 0;             /* Iteration count             */
   long        NumCycle = 0;            /* Cycle count                 */
   int         i, j;                    /* Loop Variables              */
   double      Y_Test;                  /* new merit                   */
   double      Y_Store;                 /* temporary storage           */
   int         VertexTest;              /* check whether V. in bounds  */
   double      MeanChiSquare;           /* mean of merits              */
   double      VarChiSquare;            /* standard deviation of merits*/

   float       ReflectFactor  = -1.0;   /* a reflection                */
   float       ExpandFactor   =  2.0;   /* an expansion                */
   float       ContractFactor =  0.35;  /* a contraction               */
   float       ShrinkFactor   =  0.5;   /* shrinkage                   */

   double      Yb, Yw, Ynw;             /* temp variables              */
   long        dof;                     /* degrees of freedom          */
   long        MaxIter = MAX_ITER_SIMPLEX;  /* maximum iterations      */

   size_t n      = ((struct data *) params)->n;


   /* ---------------- initialize ----------------------------------   */

   for (j = 0; j < g_nparams+1; ++j) {
     for(i = 0; i < g_nparams; ++i) {
       S[j][i] = 0.0;
     }
   }   

   FitDim = 0;
   for (j = 0; j < g_nparams; ++j) {
     if (Flags_simplex[j] == ON)  ++FitDim;
   }

   dof = n - FitDim;

   if (FitDim < 0 || dof < 0) 
     die(__LINE__, "simplex: %d parameters, %d data --> dof < 0\n", FitDim, n);

   *chi_new = -1.0;

   /* -----  Init and find the Best Vertex                 ----------- */

   SimplexInitRanges(xr, x_init);

   ErrCode = SimplexInit(Y, S, Flags_simplex, func, xr, params);
   if (ErrCode != GSL_SUCCESS)
     return(ErrCode);

   for (j = 0; j < (g_nparams+1); ++j) {       /* init to some valid index  */
     if (fabs(Y[j]) >= FLT_EPSILON) { BestChi      = Y[j]; Best = j; } 
   }
   for (j = 0; j < (g_nparams+1); ++j) {      
     if ( fabs(Y[j]) >= FLT_EPSILON && Y[j] <= BestChi ) { 
       BestChi  = Y[j]; Best = j; 
     }
   } 
   for (j = 0; j < g_nparams; ++j) {
     if (Flags_simplex[j] == ON) BestEver[j] = S[Best][j];
   }

   if (verbose) 
     printf("FitDim after Init = %5d\n", FitDim); 

   MaxIter = MAX_ITER_SIMPLEX;

   /* ------       find centre of simplex ---------------------------- */

   for (j = 0; j < g_nparams; ++j) {
      Sum = 0.0;
      for (i = 0; i < (g_nparams+1); ++i) Sum = Sum + S[i][j];
      if (Flags_simplex[j] == ON) Centre[j] = Sum / (FitDim+1);
         else Centre[j] = 0.;
   }

   /*  ---------   start iterating ----------------------------------- */

   while (1) {

     ++NumCycle;

     /* -----------  best, worst, next_worst vertices    -----------  */
  
     for (j = 0; j < (g_nparams+1); ++j) {     /* init to some valid index */
       if (fabs(Y[j]) >= FLT_EPSILON) {
	 Best      = j; 
	 Worst     = j;
	 if (j != Worst) NextWorst = j;
       }
     }

     Yb = Y[Best];
     Yw = Y[Worst];

     for (j = 0; j < (g_nparams+1); ++j) {     /* find best and worst      */ 
       if (fabs(Y[j]) >= FLT_EPSILON) { 
	 if (Y[j] <= (Yb-FLT_EPSILON))  { Best  = j; Yb = Y[j]; }
	 if (Y[j] >= (Yw+FLT_EPSILON))  { Worst = j; Yw = Y[j]; }
       }
     }

     Ynw = Y[NextWorst];
     for (j = 0; j < (g_nparams+1); ++j) {       /* find next worst        */ 
       if (fabs(Y[j]) >= FLT_EPSILON && j != Worst) {
	 if (Y[j] >= (Ynw+FLT_EPSILON)) { NextWorst = j; Ynw = Y[j]; }
       }
     }

     for (j = 0; j < g_nparams; ++j) 
       BestEver[j] = S[Best][j];
     BestChi = Y[Best];

     /* >>>>>>>>>    compute standard deviation, finish  <<<<<<<<<<< */

     MeanChiSquare = 0.0; VarChiSquare = 0.0;

     for (j = 0; j < g_nparams+1; ++j) 
       MeanChiSquare = MeanChiSquare + Y[j];
     MeanChiSquare = MeanChiSquare / (FitDim+1);

     for (j = 0; j < g_nparams+1; ++j) {
       if (fabs(Y[j]) >= FLT_EPSILON) 
	 VarChiSquare = VarChiSquare + (Y[j]-MeanChiSquare) * (Y[j]-MeanChiSquare);
     }

     VarChiSquare = sqrt(VarChiSquare / FitDim);

     if (verbose) {
       printf("\n");
       printf("-- Cycle %3ld -- Best: %10.5g     SDV (ChiSquare): %10.5g\n",
	      NumCycle, Y[Best], VarChiSquare);
     }

     if (VarChiSquare <= SimplexTol || NumCycle > MAX_ITER_SIMPLEX) 
       {

	 /* -----------  compute L.C. with best vertex --------------  */
	 
	 for (j = 0; j < g_nparams; ++j) X[j] = BestEver[j];
	 SimplexSetVariables(xr, X, Flags_simplex);
	 Y[Best] = func(xr, params);
	 
	 /* ----------------- print out -----------------------------  */
	 
	 // SimplexPrintVariables(BestChi, VarChiSquare);
	 
	 if (NumCycle > MaxIter 
	     && VarChiSquare >= (SimplexTol-FLT_EPSILON)) {
	   // fprintf(stdout, "No Convergence in Fit, Maximum Evaluation Count exceed\n");
	   *chi_new = Y[Best];
	   return(GSL_CONTINUE);
	 } else {
	   // fprintf(stdout, "Fit converged, Chi square = %g\n", Y[Best]);
	   *chi_new = Y[Best];
	   return(GSL_SUCCESS);
	 }
       }  


     /* >>>>>>>>>    the next simplex iteration         <<<<<<<<<<<< */

     ++NumIter;

     if (verbose) 
       printf("FuncEval %3ld: I reflect my worst point\n", NumIter ); 

     Y_Test  = SimplexFlow(ReflectFactor, S, Y, Centre, Worst, 
			   FitDim, &ErrCode,
			   Flags_simplex,
			   func,
			   xr, params);
     if (ErrCode == 8) return(GSL_FAILURE);

     if (Y_Test <= (Y[Best]-FLT_EPSILON)) {

       ++NumIter;

       if (verbose) 
	 printf("FuncEval %3ld: Reflection was successfull, I expand further\n",
		NumIter); 

       Y_Test  = SimplexFlow(ExpandFactor, S, Y, Centre, Worst, 
			     FitDim, &ErrCode,
			     Flags_simplex,
			     func,
			     xr, params);
       if (ErrCode != GSL_SUCCESS) 
	 return(ErrCode);
     }

     else if (Y_Test >= Y[NextWorst]) {

       ++NumIter;

       if (verbose) 
	 printf("FuncEval %3ld: No improvement, I try a contraction\n", 
		NumIter);

       Y_Store = Y[Worst];
       Y_Test  = SimplexFlow(ContractFactor, S, Y, Centre, Worst, 
			     FitDim, &ErrCode,
			     Flags_simplex,
			     func,
			     xr, params);
       if (ErrCode != GSL_SUCCESS) 
	 return(ErrCode);

       if (Y_Test >= Y_Store) {

	 if (verbose) 
	   printf("Still no improvement, I shrink towards my best point\n");

	 /* shrink the simplex                                       */
	 /* loop over vertices                                       */

	 for (i = 0; i < g_nparams+1; ++i) {
	   if (i != Best && fabs(Y[i]) >= FLT_EPSILON) {
	     /* this is not a best vertex                            */
	     /* contract and calculate anew                          */
	     ++NumIter; 
	     if (verbose) 
	       printf("    FuncEval %3ld: Computing Vertex %3d\n", 
		      NumIter,i); 
	     for (j = 0; j < g_nparams; ++j) {
	       if (Flags_simplex[j] == ON) {
		 X[j] = S[Best][j] + ShrinkFactor * (S[i][j] - S[Best][j]);
	       }
	     }
	     if (debug) { 
	       fprintf(stdout, "Best:" );
	       for (j = 0; j < g_nparams; ++j) 
		 fprintf(stdout," %7.2f", S[Best][j]);
	       fprintf(stdout,"\n");
	       fprintf(stdout, "Old:" );
	       for (j = 0; j < g_nparams; ++j) 
		 fprintf(stdout," %7.2f", S[i][j]);
	       fprintf(stdout,"\n");
	       fprintf(stdout, "New:" );
	       for (j = 0; j < g_nparams; ++j) 
		 fprintf(stdout," %7.2f", X[j]);
	       fprintf(stdout,"\n");
	     }
	     VertexTest = SimplexCheckVertex(xr, X, Flags_simplex);

	     for (j = 0; j < g_nparams; ++j) S[i][j] = X[j];

	     if (debug) {
	       for (j = 0; j < g_nparams; ++j) 
		 fprintf(stdout,"%7.2f",X[j]);
	       printf("\n");
	     }
	     if (VertexTest > 0) {
	       Y[i] = PENALTY;
	     } else {
	       SimplexSetVariables(xr, X, Flags_simplex);
	       Y[i] = func (xr, params);
	     }
	   }
	 }  /* shrink best value finished                            */


	 for (j = 0; j < g_nparams; ++j) {
	   Sum = 0.0;
	   for (i = 0; i < (g_nparams+1); ++i) Sum = Sum + S[i][j];
	   if (Flags_simplex[j] == ON) Centre[j] = Sum / (FitDim+1);
	   else Centre[j] = 0.;
	 }
       }
     }

   } /* check convergence, start new loop                            */
   *chi_new = Y[Best];
   return(GSL_SUCCESS);
}    /* END OF PROGRAM                                               */


double 
nu_of_t (double t, double e, double P, double T)
{
  double nu, sign; 
  double M   = (2.0 * 3.14159265358979323846 / P) * (t - T);
  double old, new;

  if (e >= 0.99) e = 0.99;
  if (e <  0.00) e = 0.00;


  /* normalize M to [0,2*PI] 
   */
  sign = (M >= 0) ? 1.0 : -1.0;
  M = fabs(M) / (2.0 * 3.14159265358979323846);
  M = (M - floor(M)) * (2.0 * 3.14159265358979323846) * sign;


  /* fixed point iteration: E = M + e * sin(E); 
   */
  new = M + e * sin(M);
  do {
    old = new;
    new = M + e * sin(old);
  } while ( fabs(old-new) >= 20 * DBL_EPSILON);
    
  nu = 2.0 * atan( tan(new/2.0) 
		   * ( sqrt ( (1.0 + e)
			       / (1.0 - e))));
  return nu;
}


int
orbit_f (const gsl_vector * xr, void *params, gsl_vector * f)
{
  size_t n      = ((struct data *) params)->n;
  double *x     = ((struct data *) params)->x;
  double *y     = ((struct data *) params)->y;
  double *sigma = ((struct data *) params)->sigma;
  int    *flag  = ((struct data *) params)->flag;

  double gamma  = gsl_vector_get (xr, 0);
  double K1     = gsl_vector_get (xr, 1);
  double e      = gsl_vector_get (xr, 2);
  double omega  = gsl_vector_get (xr, 3);
  double P      = gsl_vector_get (xr, 4);
  double T      = gsl_vector_get (xr, 5);
  double K2     = 1.0;
  size_t i;

  if (g_nparams == 7)
    K2     = gsl_vector_get (xr, 6);

  if (K1 <= 0.0 || e > 0.99 || e < 0.0)
    {
      for (i = 0; i < n; i++)
	gsl_vector_set (f, i, gsl_vector_get(f, i) + 100.0);
    }
  else
    {
      if (omega < 0)
	omega += 2.0 * M_PI;
      for (i = 0; i < n; i++)
	{
	  double K, Yi;
	  double nu  = nu_of_t(x[i], e, P, T);
	  
	  if (flag[i] == 1)
	    K = -(K1/K2);
	  else
	    K = K1;
	  Yi = gamma + K * ( e * cos(omega) + cos (omega + nu));
	  gsl_vector_set (f, i, ((y[i] - Yi)*(y[i] - Yi))/(sigma[i]*sigma[i]));
	}
    }
  return GSL_SUCCESS;
}

int
orbit_df (const gsl_vector * xr, void *params, gsl_matrix * df)
{
  size_t n      = ((struct data *) params)->n;
  double *x     = ((struct data *) params)->x;
  double *y     = ((struct data *) params)->y;
  double *sigma = ((struct data *) params)->sigma;
  int    *flag  = ((struct data *) params)->flag;

  double gamma  = gsl_vector_get (xr, 0);
  double K1     = gsl_vector_get (xr, 1);
  double e      = gsl_vector_get (xr, 2);
  double omega  = gsl_vector_get (xr, 3);
  double P      = gsl_vector_get (xr, 4);
  double T      = gsl_vector_get (xr, 5);
  double K2     = 1.0;

  size_t i;

  double val1, val0;
  double eps = pow(FLT_EPSILON, (1.0/3.0));
  double delta;

  if (g_nparams == 7)
    K2     = gsl_vector_get (xr, 6);

  if (omega < 0)
    omega += 2.0 * M_PI;

  for (i = 0; i < n; i++)
    {
      double K, Kd, Yi;

      double s      = sigma[i];
      double nu     = nu_of_t(x[i], e, P, T);

      if (flag[i] == 1)
	K = -(K1/K2);
      else
	K = K1;

      Yi = y[i] - (gamma + K * ( e * cos(omega) + cos (omega + nu)));
      Yi = Yi * 2.0;

      /*  Yi = gamma + K * ( e * cos(omega) + cos (omega + nu))  
       *      gsl_matrix_set (df, i, 0, -1.0/s);
       *      gsl_matrix_set (df, i, 1, -(e * cos(omega) + cos (omega + nu))/s);
       *      gsl_matrix_set (df, i, 2, -(K * cos(omega))/s);
       *      gsl_matrix_set (df, i, 3, (K * (sin(omega) + sin (omega + nu)))/s);
       *      gsl_matrix_set (df, i, 4, (K * sin (omega + nu) * dnu_dP)/s);
       *      gsl_matrix_set (df, i, 5, (K * sin (omega + nu) * dnu_dT)/s);
       */

      gsl_matrix_set (df, i, 0, Yi * (-1.0)/(s*s));

      delta = K1 * eps;
      delta = (delta < eps) ? eps : delta;
      if (flag[i] == 1)
	Kd = -((K1+delta)/K2);
      else
	Kd = K1+delta;
      val1  = y[i] - (gamma + Kd * ( e * cos(omega) + cos (omega + nu)));
      if (flag[i] == 1)
	Kd = -((K1-delta)/K2);
      else
	Kd = K1-delta;
      val0  = y[i] - (gamma + Kd * ( e * cos(omega) + cos (omega + nu)));
      
      if (debug)
	printf("%03d:  %15.8f  %15.8f  %15.8f  %15.8f\n", i, delta, val1, val0,
	       ((val1 - val0)/(2.0 * delta))/s);
      
      gsl_matrix_set (df, i, 1, Yi * ((val1 - val0)/(2.0 * delta))/(s*s));

      delta = e * eps;
      delta = (delta < eps) ? eps : delta;
      val1  = y[i] - (gamma + K * ( (e+delta) * cos(omega) + cos (omega + nu)));
      val0  = y[i] - (gamma + K * ( (e-delta) * cos(omega) + cos (omega + nu)));
      
      if (debug)
	printf("%03d:  %15.8f  %15.8f  %15.8f  %15.8f\n", i, delta, val1, val0,
	       ((val1 - val0)/(2.0 * delta))/s);
      
      gsl_matrix_set (df, i, 2, Yi * ((val1 - val0)/(2.0 * delta))/(s*s));

      delta = omega * eps;
      delta = (delta < eps) ? eps : delta;
      val1  = y[i] - (gamma + K * ( e * cos(omega+delta) + cos (omega +delta+ nu)));
      val0  = y[i] - (gamma + K * ( e * cos(omega-delta) + cos (omega -delta+ nu)));
      
      if (debug)
	printf("%03d:  %15.8f  %15.8f  %15.8f  %15.8f\n", i, delta, val1, val0,
	       ((val1 - val0)/(2.0 * delta))/s);
      
      gsl_matrix_set (df, i, 3, Yi * ((val1 - val0)/(2.0 * delta))/(s*s));

      delta = P * eps;
      delta = (delta < eps) ? eps : delta;
      nu     = nu_of_t(x[i], e, P+delta, T);
      val1  = y[i] - (gamma + K * ( e * cos(omega) + cos (omega + nu)));
      nu     = nu_of_t(x[i], e, P-delta, T);
      val0  = y[i] - (gamma + K * ( e * cos(omega) + cos (omega + nu)));
      
      if (debug)
	printf("%03d:  %15.8f  %15.8f  %15.8f  %15.8f\n", i, delta, val1, val0,
	       ((val1 - val0)/(2.0 * delta))/s);
      
      gsl_matrix_set (df, i, 4, Yi * ((val1 - val0)/(2.0 * delta))/(s*s));

      delta = T * eps;
      delta = (delta < eps) ? eps : delta;
      nu     = nu_of_t(x[i], e, P, T+delta);
      val1  = y[i] - (gamma + K * ( e * cos(omega) + cos (omega + nu)));
      nu     = nu_of_t(x[i], e, P, T-delta);
      val0  = y[i] - (gamma + K * ( e * cos(omega) + cos (omega + nu)));
      
      if (debug)
	printf("%03d:  %15.8f  %15.8f  %15.8f  %15.8f\n", i, delta, val1, val0,
	       ((val1 - val0)/(2.0 * delta))/s);
      
      gsl_matrix_set (df, i, 5, Yi * ((val1 - val0)/(2.0 * delta))/(s*s));

      if (g_nparams == 7)
	{
	  delta = K2 * eps;
	  delta = (delta < eps) ? eps : delta;
	  if (flag[i] == 1)
	    Kd = -(K1/(K2+delta));
	  else
	    Kd = K1;
	  val1  = y[i] - (gamma + Kd * ( e * cos(omega) + cos (omega + nu)));
	  if (flag[i] == 1)
	    Kd = -(K1/(K2-delta));
	  else
	    Kd = K1;
	  val0  = y[i] - (gamma + Kd * ( e * cos(omega) + cos (omega + nu)));
	  
	  if (debug)
	    printf("%03d:  %15.8f  %15.8f  %15.8f  %15.8f\n", 
		   i, delta, val1, val0,
		   ((val1 - val0)/(2.0 * delta))/s);
	  
	  gsl_matrix_set (df, i, 6, Yi * ((val1 - val0)/(2.0 * delta))/(s*s));
	}
    }

  return GSL_SUCCESS;
}

int
orbit_fdf (const gsl_vector * x, void *params, gsl_vector * f, gsl_matrix * df)
{
  orbit_f  (x, params, f);
  orbit_df (x, params, df);

  return GSL_SUCCESS;
}


double
gfunc (const gsl_vector * xr, void *params)
{
  size_t n      = ((struct data *) params)->n;
  int    *flag  = ((struct data *) params)->flag;
  double *x     = ((struct data *) params)->x;
  double *y     = ((struct data *) params)->y;
  double *sigma = ((struct data *) params)->sigma;

  double gamma  = gsl_vector_get (xr, 0);
  double K1     = gsl_vector_get (xr, 1);
  double e      = gsl_vector_get (xr, 2);
  double omega  = gsl_vector_get (xr, 3);
  double P      = gsl_vector_get (xr, 4);
  double T      = gsl_vector_get (xr, 5);
  double K2     = 1.0;

  size_t i;
  double gq = 0.0;

  if (g_nparams == 7)
    K2     = gsl_vector_get (xr, 6);

  if (omega < 0)
    omega += 2.0 * M_PI;
  for (i = 0; i < n; i++)
    {
      double nu  = nu_of_t(x[i], e, P, T);
      double K, Yi;

      if (flag[i] == 0)
	K     = K1;
      else
	K     = -(K1/K2);
      
      Yi = gamma + K * ( e * cos(omega) + cos (omega + nu));

      if (debug)
	printf("pt %d: y[i] = %8.4f   Yi = %8.4f\n", i, y[i], Yi);

      gq +=   (y[i] - Yi)*(y[i] - Yi)/(sigma[i]*sigma[i]);
    }

  gq /= n;
  if (debug)
    fprintf(stdout, "GQ: %g\n", gq);
  return gq;
}

int
print_state (size_t iter, gsl_multifit_fdfsolver * s, int p)
{
  int j;

  printf ("iter: %3u x = ", iter);
  for (j = 0; j < p; ++j)
    printf ("%8.4f ", gsl_vector_get (s->x, j));
  printf ("|f(x)| = %g\n", gsl_blas_dnrm2 (s->f));

  return 0;
}

int
print_simplex_state (size_t iter, const gsl_vector * x, double new, int p)
{
  int j;

  printf ("iter: %3u x = ", iter);
  for (j = 0; j < p; ++j)
    printf ("%8.4f ", gsl_vector_get (x, j));
  printf ("|f(x)| = %g\n", new);

  return 0;
}



int
main (int argc, char * argv[])
{
  int tflag = 0;  /* test only    */
  int fflag = 0;  /* fixed params */
  int pflag = 0;  /* user params  */

  double  old = 1.0, new;

  int quiet = 0;
  int wfit  = 1;

  const gsl_multifit_fdfsolver_type *T;
  gsl_multifit_fdfsolver *s;

  int status;
  size_t i, iter = 0;

  size_t n = 60;
  size_t p = 6;

  gsl_matrix * covar;

  double * y, * xi, * sigma, ymin, ymax; 
  int    * flag;

  struct data d;
  
  gsl_multifit_function_fdf f; 

  double * x_init;
  double x_init_1[6] = { 0.0, 1.0, 0.0, 0.5, 1.0, 0.0};
  double x_init_2[7] = { 0.0, 1.0, 0.0, 0.5, 1.0, 0.0, 1.0};

  int    Flags_simplex[7] = { ON, ON, ON, ON, ON, ON, ON };

  gsl_vector_view x;

  const gsl_rng_type * mT;
  gsl_rng * r;

  /* setup random number generator
   */
  gsl_rng_env_setup();
  mT = gsl_rng_default;
  r  = gsl_rng_alloc (mT);

  /* command line
   */
  i = 1;

  do {
    if (argc <= i)
      break;
    if (argv[i][0] != '-')
      break;
    
    switch (argv[i][1]) {
      
    case 'n':
      if (argc <= i+1)
	die (__LINE__, "-n without parameters\n");
      n = atof(argv[i+1]);
      if (n < 7 || n == HUGE_VAL || n == (-HUGE_VAL))
	die (__LINE__, "-n with bad argument %s\n", argv[i+1]);
      ++i;
      break;
      
    case 'L':
      wfit = 2;
      break;
      
    case 'v':
      verbose = 1;
      break;
      
    case 'q':
      quiet = 1;
      break;
      
    case 'f':
      fflag = 1;
      break;
      
    case 'd':
      debug = 1;
      break;
      
    case 'P':
      if (argc <= i+1)
	die (__LINE__, "-P without parameters\n");
      Device = argv[i+1];
      ++i;
      break;
      
    case 'c':
      {
	char * p;
	int    j, a;
	if (argc <= i+1)
	  die (__LINE__, "-c without parameters\n");
	p = strtok(argv[i+1], ",");
	if (!p)
	  die (__LINE__, "-c without parameters\n");
	a = atoi(p);
	if (a != 0 && a != 1)
	  die (__LINE__, "-c with invalid argument %s\n", p);
	Flags_simplex[0] = a;
	for (j = 1; j < 7; ++j)
	  {
	    p = strtok(NULL, ",");
	    if (!p)
	      die (__LINE__, "-c with too few parameters\n");
	    a = atoi(p);
	    if (a != 0 && a != 1)
	      die (__LINE__, "-c with invalid argument %s\n", p);
	    Flags_simplex[j] = a;
	  }
	++i;
	break;
      }
      
    case 'p':
      {
	char * p;
	int    j;
	if (argc <= i+1)
	  die (__LINE__, "-p without parameters\n");
	p = strtok(argv[i+1], ",");
	if (!p)
	  die (__LINE__, "-p without parameters\n");
	x_init_2[0] = atof(p);
	for (j = 1; j < 7; ++j)
	  {
	    p = strtok(NULL, ",");
	    if (!p)
	      die (__LINE__, "-p with too few parameters\n");
	    x_init_2[j] = atof(p);
	  }
	pflag = 1;
	for (j = 0; j < 6; ++j)
	  x_init_1[j] = x_init_2[j];

	for (j = 0; j < 7; ++j)
	  {
	    if (x_init_2[j] < lim_low[j] || x_init_2[j] > lim_hig[j]) {
	      fprintf(stderr, 
		      "*** -p: Out of range [%8.4g,%8.4g]: %8s, Value = %8.4g\n",
		      lim_low[j],
		      lim_hig[j],
		      Param[j].Name,
		      x_init_2[j]);
	      out_of_range();
	      die(__LINE__, "-p with invalid argument %8.4g\n", x_init_2[j]);
	    }
	  }

	++i;
	break;
      }
      
    case '-':
    case 'h':
    default:
      printf("Usage:   %s [options] datafile\n",               argv[0]);
      printf("         %s [options] testonly_1 (test SB1)\n",  argv[0]);
      printf("         %s [options] testonly_2 (test SB2)\n",  argv[0]);
      printf("         %s -h\n",                               argv[0]);
      printf("\n");
      printf("Options: -n N  number of data points for 'testonly' (60)\n");
      printf("         -p gamma,K1,e,omega,P,T,[K1/K2]\n");
      printf("               start parameters for fit, use [K1/K2]=1.0 for SB1\n");
      printf("               (0.0,1.0,0.0,0.5,1.0,0.0,1.0)\n");
      printf("         -c (0|1),(0|1),(0|1),(0|1),(0|1),(0|1),(0|1)\n");
      printf("               parameters to hold constant (0) or fit (1)\n");
      printf("               (1,1,1,1,1,1,1)\n");
      printf("         -P device_name\n");
      printf("               output device name for PGPLOT (/XSERVE, use /PS for hardcopy)\n");
      printf("         -L    plot Levenberg-Marquardt fit instead of Simplex fit\n");
      printf("         -f    parameters are fixed\n");
      printf("         -q    be more quiet\n");
      printf("         -v    be more verbose\n");
      printf("         -d    debug output (very verbose)\n");
      printf("\n");
      printf("Datafile format: ('testonly' generates random data internally)\n");
      printf("         - comment lines start with a '#'\n");
      printf("         - each data line: 4 numbers, seperated by blank space\n");
      printf("         time velocity sigma flag\n");
      printf("         - flag must be 0 for SB1, else 0 for primary, 1 for secondary\n");
      exit (0);
    }
    ++i;
  } while (1 == 1);


  if (argc < i+1)
    die (__LINE__, "Mandatory argument misssing (use '%s -h' for help)\n",
	 argv[0]);


  if (0 == strcmp(argv[i], "testonly_1"))
    {
      tflag   = 1;
      p       = 6;  /* test SB1  */
      xi      = (double *) xalloc (2 * n * sizeof(double));
      y       = (double *) xalloc (2 * n * sizeof(double));
      sigma   = (double *) xalloc (2 * n * sizeof(double));
      flag    = (int *)    xalloc (2 * n * sizeof(int));
      
      d.n     = n;
      d.x     = xi;
      d.y     = y;
      d.sigma = sigma;
      d.flag  = flag;
    }
  else if (0 == strcmp(argv[i], "testonly_2"))
    {
      tflag   = 1;
      p       = 7;  /* always test SB2  */
      xi      = (double *) xalloc (2 * n * sizeof(double));
      y       = (double *) xalloc (2 * n * sizeof(double));
      sigma   = (double *) xalloc (2 * n * sizeof(double));
      flag    = (int *)    xalloc (2 * n * sizeof(int));
      
      d.n     = 2 * n;
      d.x     = xi;
      d.y     = y;
      d.sigma = sigma;
      d.flag  = flag;
    }
  else 
    {
      FILE * fd = fopen(argv[i], "r");
      char line[128];
      size_t count = 0;
      size_t nsize  = 128;

      double ax, ay, as;
      int    af;
      
      xi      = (double *) xalloc (nsize * sizeof(double));
      y       = (double *) xalloc (nsize * sizeof(double));
      sigma   = (double *) xalloc (nsize * sizeof(double));
      flag    = (int *)    xalloc (nsize * sizeof(int));
      
      if (!fd)
	die(__LINE__, "Cannot open %s for reading\n", argv[i]);
      while (NULL != fgets(line, 127, fd))
	{
	  if (line[0] == '#')
	    continue;
	  if (4 == sscanf(line, "%lf %lf %lf %d\n", &ax, &ay, &as, &af))
	    {
	      if (ax < lim_low[5] || ax > lim_hig[5]) {
		fprintf(stdout, 
			"*** %s: Out of range [%6.1g,%6.1g]: %s, Value = %8.4g\n",
			argv[i],
			lim_low[5],
			lim_hig[5],
			"x (date)",
			ax);
		out_of_range();
	      } else if (fabs(ay) > lim_hig[1]) {
		fprintf(stdout, 
			"*** %s: Out of range [%6.1g,%6.1g]: %s, Value = %8.4g\n",
			argv[i],
			lim_low[1],
			lim_hig[1],
			"y (velocity)",
			ay);
		out_of_range();
	      } else {
		xi[count]    = ax;
		y[count]     = ay;
		sigma[count] = as;
		flag[count]  = af;
		if (af == 1)
		  p = 7;
		++count;
	      }
	    }
	  else
	    {
	      int j = 0;
	      do {
		if (j == 128 || line[j] == '\0')
		  break;
		if (line[j] != ' ' && line[j] != '\t' && line[j] != '#')
		  {
		    printf("*** WARN **: %s: Invalid line: %s\n", argv[i], line);
		    printf("*** WARN **: required format: time velocity sigma flag\n");
		    break;
		  }
		++j;
	      } while (j < 128 && line[j] != '\0');
	    }
	  if (count == nsize)
	    {
	      nsize += 128;
	      xi     = (double *) realloc (xi,    nsize * sizeof(double));
	      y      = (double *) realloc (y,     nsize * sizeof(double));
	      sigma  = (double *) realloc (sigma, nsize * sizeof(double));
	      flag   = (int *)    realloc (flag,  nsize * sizeof(int));
	    }
	}
      fclose(fd);
      
      n       = count;
      d.n     = n;
      d.x     = xi;
      d.y     = y;
      d.sigma = sigma;
      d.flag  = flag;
    }

  /* Test only -- generate test data
   */
  if (tflag)
    {
      double gamma, K1, e, omega, P, T, K2, t, nu;
      if (pflag)
	{
	  gamma = x_init_2[0] + gsl_ran_gaussian(r, 0.5 * x_init_2[0]);
	  K1    = x_init_2[1] + gsl_ran_gaussian(r, 0.5 * x_init_2[1]);
	  e     = x_init_2[2] + gsl_ran_gaussian(r, 0.5 * x_init_2[2]);
	  omega = x_init_2[3] + gsl_ran_gaussian(r, 0.5 * x_init_2[3]);
	  P     = x_init_2[4] + gsl_ran_gaussian(r, 0.5 * x_init_2[4]);
	  T     = x_init_2[5] + gsl_ran_gaussian(r, 0.5 * x_init_2[5]);
	  K2    = x_init_2[6] + gsl_ran_gaussian(r, 0.5 * x_init_2[6]);
	}
      else
	{
	  int j;

	  gamma  = 3.0;      x_init_2[0] = gamma;
	  K1     = 25.50;    x_init_2[1] = K1;
	  e      = 0.3;      x_init_2[2] = e;
	  omega  = M_PI/4.0; x_init_2[3] = omega;
	  P      = 12.4;     x_init_2[4] = P;
	  T      = 0.0;      x_init_2[5] = T;
	  if (p == 6)
	    K2 = 1.0;
	  else
	    K2     = 1.2;
	   x_init_2[6] = K2;
	   for (j = 0; j < 6; ++j)
	     x_init_1[j] = x_init_2[j];
	}

      printf("\n--- Test run, orbit parameters                  ---\n");
      printf("gamma  = %g\n", gamma);
      printf("K1     = %g\n", K1);
      printf("e      = %g\n", e);
      printf("omega  = %g\n", omega);
      printf("P      = %g\n", P);
      printf("T      = %g\n", T);
      printf("K1/K2  = %g\n", K2);

      for (i = 0; i < n; i++)
	{
	  t  = T + 2.0 * (P/n) * i;           /* time */
	  nu = nu_of_t (t, e, P, T);
	  
	  xi[i]      = t;
	  if (p == 7)
	    xi[i+n]    = t;
	  y[i]       = gamma + K1       * ( e * cos(omega) + cos(omega + nu) ) 
	    + gsl_ran_gaussian(r, 0.05 * K1);
	  if (p == 7)
	    y[i+n]     = gamma + (-K1/K2) * ( e * cos(omega) + cos(omega + nu) ) 
	      + gsl_ran_gaussian(r, 0.05 * K1);
	  sigma[i]   = 0.05 * K1;
	  if (p == 7)
	    sigma[i+n] = 0.05 * K1;
	  flag[i]    = 0;
	  if (p == 7)
	    flag[i+n]  = 1;
	}

      if (p == 7)
	n = 2 * n;
    }

  if (n <= 0)
    die(__LINE__, "No valid input data.\n");
      
  /* printf("FIXME (p)     = %d\n", p); */
  g_nparams = p;

  if (p == 6)
    x_init = &x_init_1[0];
  else
    x_init = &x_init_2[0];
      
  if (!pflag)
    {
      ymin = y[0];
      ymax = y[0];
      for (i = 0; i < n; i++)
	{
	  ymin = GSL_MIN(ymin, y[i]);
	  ymax = GSL_MAX(ymax, y[i]);
	}
      x_init[0] = ymin + (ymax - ymin)/2.0;
      x_init[1] = (ymax - ymin)/2.0;
    }

#define FIT(i)  gsl_vector_get(s->x, i)
#define SFIT(i) gsl_vector_get(&x.vector, i)
#define ERR(i)  sqrt(gsl_matrix_get(covar,i,i))

  /* iterate
   */
  f.f        = &orbit_f;
  f.df       = &orbit_df;
  f.fdf      = &orbit_fdf;
  f.n        = n;
  f.p        = p;
  f.params   = (void *) &d;
 
  T = gsl_multifit_fdfsolver_lmsder;
  x = gsl_vector_view_array (x_init, p); 
  s = gsl_multifit_fdfsolver_alloc (T, n, p);

  status = GSL_EFAILED;

  if (!fflag)
    {
      int j;

      for (j = 0; j < 200; ++j)
	{
	  status = Simplex( x_init,  Flags_simplex, &new,
			    gfunc, &x.vector, (void *) &d );
	  if (status == GSL_FAILURE)
	    break;
	  if (fabs(old-new) < (10.0 * FLT_EPSILON))
	    break;
	  if (!quiet)
	    {
	      print_simplex_state (j, &x.vector, new, p);
	      printf ("status = %s\n", gsl_strerror (status));
	    }

	  old = new;
	  for (i = 0; i < p; i++)
	    {
	      if (Flags_simplex[i] == ON)
		x_init[i] = gsl_vector_get(&x.vector, i);
	    }
	}

      printf("\n ------  After %d iterations of simplex\n\n", j);
      printf(" gamma  = %g\n", SFIT(0));
      printf(" K1     = %g\n", SFIT(1));
      printf(" e      = %g\n", SFIT(2));
      printf(" omega  = %g\n", SFIT(3));
      printf(" P      = %g\n", SFIT(4));
      printf(" T      = %g\n", SFIT(5));
      if (p == 7)
	printf(" K1/K2  = %g\n", SFIT(6));
      else
	printf(" K1/K2  = (n/a)\n");

      if (p == 7)
	printf("\n To restart, use: -p %g,%g,%g,%g,%g,%g,%g\n", 
	       SFIT(0), SFIT(1), SFIT(2), SFIT(3), SFIT(4), SFIT(5), SFIT(6) );
      else
	printf("\n To restart, use: -p %g,%g,%g,%g,%g,%g,0.0\n", 
	       SFIT(0), SFIT(1), SFIT(2), SFIT(3), SFIT(4), SFIT(5));

      printf (" status = %s\n", gsl_strerror (status));
    }

  if (wfit == 1)
    plot_result (SFIT(0), SFIT(1), SFIT(2), SFIT(3), SFIT(4), SFIT(5), 
		 (p == 7) ? SFIT(6) : 0.0, 
		 n, xi, y, sigma, flag);

  plot_statistics (SFIT(0), SFIT(1), SFIT(2), SFIT(3), SFIT(4), SFIT(5), 
		   (p == 7) ? SFIT(6) : 0.0, 
		   n, xi, y, sigma, flag);


  gsl_multifit_fdfsolver_set (s, &f, &x.vector);
  if (!quiet)
    print_state (iter, s, p);

  covar = gsl_matrix_alloc (p, p);
  gsl_multifit_covar (s->J, 0.0, covar);

  printf("\n ------  After Levenberg-Marquardt setup\n\n");
  printf(" gamma  = %g +/- %g\n", FIT(0), ERR(0));
  printf(" K1     = %g +/- %g\n", FIT(1), ERR(1));
  printf(" e      = %g +/- %g\n", FIT(2), ERR(2));
  printf(" omega  = %g +/- %g\n", FIT(3), ERR(3));
  printf(" P      = %g +/- %g\n", FIT(4), ERR(4));
  printf(" T      = %g +/- %g\n", FIT(5), ERR(5));
  if (p == 7)
    printf(" K1/K2  = %g +/- %g\n", FIT(6), ERR(6));
  else
    printf(" K1/K2  = (n/a)\n");


  if (fflag)
    exit (0);
  
  do
    {
      iter++;
      status = gsl_multifit_fdfsolver_iterate (s);
      if (!quiet)
	{
	  printf ("status = %s\n", gsl_strerror (status));
	  print_state (iter, s, p);
	}
      if (status && status != GSL_CONTINUE)
        break;
      status = gsl_multifit_test_delta (s->dx, s->x, FLT_EPSILON, FLT_EPSILON);
    }
  while (status == GSL_CONTINUE && iter < 500);

  gsl_multifit_covar (s->J, 0.0, covar);

  printf("\n ------  After Levenberg-Marquardt (%3d iterations)\n\n", iter);
  printf(" gamma  = %g +/- %g\n", FIT(0), ERR(0));
  printf(" K1     = %g +/- %g\n", FIT(1), ERR(1));
  printf(" e      = %g +/- %g\n", FIT(2), ERR(2));
  printf(" omega  = %g +/- %g\n", FIT(3), ERR(3));
  printf(" P      = %g +/- %g\n", FIT(4), ERR(4));
  printf(" T      = %g +/- %g\n", FIT(5), ERR(5));
  if (p == 7)
    printf(" K1/K2  = %g +/- %g\n", FIT(6), ERR(6));
  else
    printf(" K1/K2  = (n/a)\n");
  printf (" status = %s\n", gsl_strerror (status));

  if (wfit == 2)
    plot_result (FIT(0), FIT(1), FIT(2), FIT(3), FIT(4), FIT(5), 
		 (p == 7) ? FIT(6) : 0.0, 
		 n, xi, y, sigma, flag);

  plot_statistics (FIT(0), FIT(1), FIT(2), FIT(3), FIT(4), FIT(5), 
		   (p == 7) ? FIT(6) : 0.0, 
		   n, xi, y, sigma, flag);

  gsl_multifit_fdfsolver_free (s);

  return 0;
}


int
plot_result (float gamma, float K, float e, float omega, float P, float T, float K2,
	     int dn, double *dxi, double *dyi, double *dsi, int *dfi)
{
  float x[100], y[100];
  float *dx, *dy, *ds;
  float *dx2, *dy2, *ds2;
  int   di = 0, di2 = 0;

  float xmin, xmax, ymin, ymax, xdelta, ydelta;
  int   i;

  dx =  (float*) xalloc (dn * sizeof(float));
  dy =  (float*) xalloc (dn * sizeof(float));
  ds =  (float*) xalloc (dn * sizeof(float));
  dx2 = (float*) xalloc (dn * sizeof(float));
  dy2 = (float*) xalloc (dn * sizeof(float));
  ds2 = (float*) xalloc (dn * sizeof(float));

  for (i = 0; i < dn; ++i)
    {
      if (dfi[i] == 0)
	{
	  dx[di] = dxi[i];
	  dy[di] = dyi[i];
	  ds[di] = dsi[i];
	  ++di;
	}
      else
	{
	  dx2[di2] = dxi[i];
	  dy2[di2] = dyi[i];
	  ds2[di2] = dsi[i];
	  ++di2;
	}
    }


  if(cpgbeg(0, Device, 1, 1) != 1)
    die (__LINE__, "Cannot open device %s\n", Device);

  xmin = dxi[0];
  xmax = dxi[0];
  ymin = dyi[0];
  ymax = dyi[0];

  for (i = 1; i < dn; ++i)
    {
      xmin = GSL_MIN(xmin, dxi[i]);
      xmax = GSL_MAX(xmax, dxi[i]);
      ymin = GSL_MIN(ymin, dyi[i]);
      ymax = GSL_MAX(ymax, dyi[i]);
    }

  xdelta = (xmax - xmin);
  ydelta = (ymax - ymin);
      

  cpgsch(1.5);
  cpgscf(2);

  cpgenv(xmin - (xdelta/dn), xmax + (xdelta/dn), 
	 ymin - (ydelta/dn), ymax + (ydelta/dn), 
	 0, 1);

  cpglab("Date", "V\\drad\\u [km/s]", "");

  if (di > 0)
    cpgpt(di,  dx,  dy,  4);
  if (di2 > 0)
    cpgpt(di2, dx2, dy2, -8);

  for (i = 0; i < 100; ++i)
    {
      double nu;
      x[i] = (xmin + i * xdelta / 100);
      nu   = nu_of_t (x[i], e, P, T);
      y[i] = gamma + K * ( e * cos(omega) + cos (omega + nu));
    }
  cpgline(100, x, y);

  if (K2 > FLT_EPSILON)
    {
      K = -(K/K2);
      for (i = 0; i < 100; ++i)
	{
	  double nu;
	  x[i] = (xmin + i * xdelta / 100);
	  nu   = nu_of_t (x[i], e, P, T);
	  y[i] = gamma + K * ( e * cos(omega) + cos (omega + nu));
	}
      cpgline(100, x, y);
    }

  cpgend();
  return 0;
}

int
plot_statistics (float gamma, float K, float e, float omega, 
		 float P, float T, float K2,
		 int dn, double *dxi, double *dyi, double *dsi, int *dfi)
{
  float rms = 0, rms2 = 0, chi2 = 0, tt;
  float *dx, *dy, *ds, *derr;
  float *dx2, *dy2, *ds2, *derr2;
  int   di = 0, di2 = 0;
  int   i;

  dx =  (float*) xalloc (dn * sizeof(float));
  dy =  (float*) xalloc (dn * sizeof(float));
  ds =  (float*) xalloc (dn * sizeof(float));
  dx2 = (float*) xalloc (dn * sizeof(float));
  dy2 = (float*) xalloc (dn * sizeof(float));
  ds2 = (float*) xalloc (dn * sizeof(float));

  derr  = (float*) xalloc (dn * sizeof(float));
  derr2 = (float*) xalloc (dn * sizeof(float));

  for (i = 0; i < dn; ++i)
    {
      if (dfi[i] == 0) /* primary/secondary flag */
	{
	  dx[di] = dxi[i];
	  dy[di] = dyi[i];
	  ds[di] = dsi[i];
	  derr[di] = 0;
	  ++di;
	}
      else
	{
	  dx2[di2] = dxi[i];
	  dy2[di2] = dyi[i];
	  ds2[di2] = dsi[i];
	  derr2[di] = 0;
	  ++di2;
	}
    }

  for (i = 0; i < di; ++i)
    {
      double nu;
      nu   = nu_of_t (dx[i], e, P, T);
      derr[i] = gamma + K * ( e * cos(omega) + cos (omega + nu));

      tt    = (derr[i] - dy[i]) * (derr[i] - dy[i]);
      rms  += tt;
      chi2 += (tt/ds[i]);
    }
  if (di > 1)
    rms /= (di -1);
  rms = sqrt(rms);

  printf ("\n ----------------------------------------------\n\n");
  printf (" RMS (Primary) = %f\n", rms);

  if (K2 > FLT_EPSILON)
    {
      K = -(K/K2);
      for (i = 0; i < di2; ++i)
	{
	  double nu;
	  nu   = nu_of_t (dx2[i], e, P, T);
	  derr2[i] = gamma + K * ( e * cos(omega) + cos (omega + nu));

	  tt    = (derr2[i] - dy2[i]) * (derr2[i] - dy2[i]);
	  rms2 += tt;
	  chi2 += (tt/ds2[i]);
	}
      if (di2 > 1)
	rms2 /= (di2 -1);
      rms2 = sqrt(rms2);
      printf (" RMS (Secondary) = %f\n", rms2);

      printf (" RMS (Total) = %f\n\n", sqrt(0.5 * (rms2 * rms2 + rms * rms)));
      printf (" CHI^2       = %f (data points: %d)\n\n", chi2, dn);
    }
  return 0;
}

